# coding: utf-8
# Copyright (c) Jan Janssen

import os
import pandas
import unittest
import getpass
from pysqa import QueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


df_queue_status = pandas.DataFrame(
    {
        "jobid": [5322019, 5322016, 5322017, 5322018, 5322013],
        "user": ["janj", "janj", "janj", "janj", "maxi"],
        "jobname": ["pi_19576488", "pi_19576485", "pi_19576486", "pi_19576487", "pi_19576482"],
        "status": ["running", "running", "running", "running", "running"],
        "working_directory": [
            "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_1",
            "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_2",
            "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_3",
            "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_4",
            "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_5",
        ]
    }
)


class TestSlurmQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.slurm = QueueAdapter(directory=os.path.join(cls.path, "config/slurm"))

    def test_config(self):
        self.assertEqual(self.slurm.config["queue_type"], "SLURM")
        self.assertEqual(self.slurm.config["queue_primary"], "slurm")

    def test_list_clusters(self):
        self.assertEqual(self.slurm.list_clusters(), ['default'])

    def test_remote_flag(self):
        self.assertFalse(self.slurm._adapter.remote_flag)

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.slurm.ssh_delete_file_on_remote, True)

    def test_interfaces(self):
        self.assertEqual(
            self.slurm._adapter._commands.submit_job_command, ["sbatch", "--parsable"]
        )
        self.assertEqual(self.slurm._adapter._commands.delete_job_command, ["scancel"])
        self.assertEqual(
            self.slurm._adapter._commands.get_queue_status_command,
            ["squeue", "--format", "%A|%u|%t|%.15j|%Z", "--noheader"],
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("slurm"):
            self.assertEqual(
                self.slurm._adapter._list_command_to_be_executed(None, "here"),
                ["sbatch", "--parsable", "here"],
            )
        with self.subTest("slurm with one dependency"):
            self.assertEqual(
                self.slurm._adapter._list_command_to_be_executed(["2"], "here"),
                ["sbatch", "--parsable", "--dependency=afterok:2", "here"],
            )
        with self.subTest("slurm with two dependencies"):
            self.assertEqual(
                self.slurm._adapter._list_command_to_be_executed(["2", "34"], "here"),
                ["sbatch", "--parsable", "--dependency=afterok:2,34", "here"],
            )

    def test_convert_queue_status_slurm(self):
        with open(os.path.join(self.path, "config/slurm", "squeue_output"), "r") as f:
            content = f.read()
        self.assertTrue(
            df_queue_status.equals(
                self.slurm._adapter._commands.convert_queue_status(
                    queue_status_output=content
                )
            )
        )

    def test_get_user(self):
        self.assertEqual(self.slurm._adapter._get_user(), getpass.getuser())

    def test_queue_view(self):
        self.assertIsInstance(self.slurm.queue_view, pandas.DataFrame)

    def test_submit_job_empty_working_directory(self):
        with self.assertRaises(ValueError):
            self.slurm.submit_job(working_directory=" ")

    def test_write_queue(self):
        with self.assertRaises(ValueError):
            self.slurm._adapter._write_queue_script(
                queue=None,
                job_name=None,
                working_directory=None,
                cores=None,
                memory_max=None,
                run_time_max=None,
                command=None
            )
        self.slurm._adapter._write_queue_script(
            queue="slurm",
            job_name=None,
            working_directory=None,
            cores=None,
            memory_max=None,
            run_time_max=None,
            command="echo \"hello\""
        )
        with open("run_queue.sh", "r") as f:
            content = f.read()
        output = """\
#!/bin/bash
#SBATCH --output=time.out
#SBATCH --job-name=None
#SBATCH --chdir=.
#SBATCH --get-user-env=L
#SBATCH --partition=slurm
#SBATCH --time=4320
#SBATCH --cpus-per-task=10

echo \"hello\""""
        self.assertEqual(content, output)
        os.remove("run_queue.sh")

    def test_write_queue_extra_keywords(self):
        self.slurm._adapter._write_queue_script(
            queue="slurm_extra",
            job_name=None,
            working_directory=None,
            cores=None,
            memory_max=None,
            run_time_max=None,
            command="echo \"hello\"",
            account="123456"
        )
        with open("run_queue.sh", "r") as f:
            content = f.read()
        output = """\
#!/bin/bash
#SBATCH --output=time.out
#SBATCH --job-name=None
#SBATCH --chdir=.
#SBATCH --get-user-env=L
#SBATCH --partition=slurm
#SBATCH --account=123456
#SBATCH --time=4320
#SBATCH --cpus-per-task=10

echo \"hello\""""
        self.assertEqual(content, output)
        os.remove("run_queue.sh")

    def test_no_queue_id_returned(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            pass

        slurm_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/slurm"),
            execute_command=execute_command
        )
        self.assertIsNone(slurm_tmp.submit_job(
            queue="slurm",
            job_name="test",
            working_directory=".",
            command="echo hello"
        ))
        self.assertIsNone(slurm_tmp.delete_job(process_id=123))

    def test_queue_status(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            with open(os.path.join(self.path, "config", "slurm", "squeue_output")) as f:
                return f.read()

        slurm_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/slurm"),
            execute_command=execute_command
        )
        self.assertTrue(
            df_queue_status.equals(slurm_tmp.get_queue_status())
        )
        self.assertTrue(
            df_queue_status[df_queue_status.user=="janj"].equals(slurm_tmp.get_queue_status(user="janj"))
        )
        self.assertEqual(slurm_tmp.get_status_of_job(process_id=5322019), "running")
        self.assertIsNone(slurm_tmp.get_status_of_job(process_id=0))
        self.assertEqual(slurm_tmp.get_status_of_jobs(process_id_lst=[5322019, 0]), ["running", "finished"])

    def test_not_implemented_functions(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            pass

        slurm_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/slurm"),
            execute_command=execute_command
        )

        with self.assertRaises(NotImplementedError):
            slurm_tmp._adapter.convert_path_to_remote(path="test")

        with self.assertRaises(NotImplementedError):
            slurm_tmp._adapter.transfer_file(file="test", transfer_back=False)

        with self.assertRaises(NotImplementedError):
            slurm_tmp._adapter.get_job_from_remote(working_directory=".")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pysqa.wrapper.generic import SchedulerCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Feb 9, 2019"


class MoabCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["msub"]

    @property
    def delete_job_command(self):
        return ["mjobctl", "-c"]

    @property
    def get_queue_status_command(self):
        return ["mdiag", "-x"]



import yaml


def read_config(file_name="queue.yaml"):
    """

    Args:
        file_name (str):

    Returns:
        dict:
    """
    with open(file_name, "r") as f:
        return yaml.load(f, Loader=yaml.FullLoader)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from abc import ABC, abstractmethod

__author__ = "Niklas Siemer"
__copyright__ = (
    "Copyright 2022, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Niklas Siemer"
__email__ = "siemer@mpie.de"
__status__ = "production"
__date__ = "Aug 15, 2022"


class SchedulerCommands(ABC):
    @property
    @abstractmethod
    def submit_job_command(self):
        pass

    @property
    @abstractmethod
    def delete_job_command(self):
        pass

    @property
    def enable_reservation_command(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def get_queue_status_command(self):
        pass

    @staticmethod
    def dependencies(dependency_list) -> list:
        if dependency_list is not None:
            raise NotImplementedError()
        else:
            return []

    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        raise NotImplementedError()

    @staticmethod
    def convert_queue_status(queue_status_output):
        raise NotImplementedError()



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest
from pysqa import QueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


class TestMoabQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.moab = QueueAdapter(directory=os.path.join(cls.path, "config/moab"))

    def test_config(self):
        self.assertEqual(self.moab.config["queue_type"], "MOAB")
        self.assertEqual(self.moab.config["queue_primary"], "moab")

    def test_list_clusters(self):
        self.assertEqual(self.moab.list_clusters(), ['default'])

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.moab.ssh_delete_file_on_remote, True)

    def test_interfaces(self):
        self.assertEqual(self.moab._adapter._commands.submit_job_command, ["msub"])
        self.assertEqual(
            self.moab._adapter._commands.delete_job_command, ["mjobctl", "-c"]
        )
        self.assertEqual(
            self.moab._adapter._commands.get_queue_status_command, ["mdiag", "-x"]
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("moab with dependency"):
            self.assertRaises(
                NotImplementedError,
                self.moab._adapter._list_command_to_be_executed,
                [],
                "here",
            )
        with self.subTest("moab"):
            self.assertEqual(
                self.moab._adapter._list_command_to_be_executed(None, "here"),
                ["msub", "here"],
            )



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import pandas
import unittest
from pysqa import QueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


class TestTorqueQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.torque = QueueAdapter(directory=os.path.join(cls.path, "config/torque"))

    def test_config(self):
        self.assertEqual(self.torque.config["queue_type"], "TORQUE")
        self.assertEqual(self.torque.config["queue_primary"], "torque")

    def test_list_clusters(self):
        self.assertEqual(self.torque.list_clusters(), ['default'])

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.torque.ssh_delete_file_on_remote, True)

    def test_interfaces(self):
        self.assertEqual(
            self.torque._adapter._commands.submit_job_command, ["qsub"]
        )
        self.assertEqual(self.torque._adapter._commands.delete_job_command, ["qdel"])
        self.assertEqual(
            self.torque._adapter._commands.get_queue_status_command, ["qstat", "-f"]
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("torque"):
            self.assertEqual(
                self.torque._adapter._list_command_to_be_executed(None, "here"),
                ["qsub", "here"],
            )
        with self.subTest("torque with dependency"):
            self.assertRaises(
                NotImplementedError,
                self.torque._adapter._list_command_to_be_executed,
                [],
                "here",
            )

    def test_convert_queue_status_torque(self):
        with open(os.path.join(self.path, "config/torque", "PBSPro_qsub_output"), "r") as f:
            content = f.read()
        df_verify = pandas.DataFrame(
            {
                "jobid": [80005196, 80005197, 80005198],
                "user": ["asd562", "asd562", "fgh562"],
                "jobname": ["test1", "test2", "test_asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf"],
                "status": ["running", "pending", "pending"],
                "working_directory": ["/scratch/a01/asd562/VASP/test/test1", \
                                      "/scratch/a01/asd562/VASP/test/test2", \
                                      "/scratch/a01/fgh562/VASP/test/test_asdfasdfasdfasdfasdfasdfasdfasdfasdfasdf"]
            }
        )
        self.assertTrue(
            df_verify.equals(
                self.torque._adapter._commands.convert_queue_status(
                    queue_status_output=content
                )
            )
        )



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import defusedxml.ElementTree as ETree
import pandas

from pysqa.wrapper.generic import SchedulerCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


class SunGridEngineCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["qsub", "-terse"]

    @property
    def delete_job_command(self):
        return ["qdel"]

    @property
    def enable_reservation_command(self):
        return ["qalter", "-R", "y"]

    @property
    def get_queue_status_command(self):
        return ["qstat", "-xml"]

    @staticmethod
    def convert_queue_status(queue_status_output):
        def leaf_to_dict(leaf):
            return [
                {sub_child.tag: sub_child.text for sub_child in child} for child in leaf
            ]

        tree = ETree.fromstring(queue_status_output)
        df_running_jobs = pandas.DataFrame(leaf_to_dict(leaf=tree[0]))
        df_pending_jobs = pandas.DataFrame(leaf_to_dict(leaf=tree[1]))
        df_merge = pandas.concat([df_running_jobs, df_pending_jobs], sort=True)
        df_merge.loc[df_merge.state == "r", "state"] = "running"
        df_merge.loc[df_merge.state == "qw", "state"] = "pending"
        df_merge.loc[df_merge.state == "Eqw", "state"] = "error"
        return pandas.DataFrame(
            {
                "jobid": pandas.to_numeric(df_merge.JB_job_number),
                "user": df_merge.JB_owner,
                "jobname": df_merge.JB_name,
                "status": df_merge.state,
                "working_directory": [""] * len(df_merge),
            }
        )



# coding: utf-8
# Copyright (c) Jan Janssen

import json
import os
import warnings

import pandas
import paramiko
from tqdm import tqdm

from pysqa.utils.basic import BasisQueueAdapter
from pysqa.utils.execute import execute_command


class RemoteQueueAdapter(BasisQueueAdapter):
    def __init__(self, config, directory="~/.queues", execute_command=execute_command):
        super(RemoteQueueAdapter, self).__init__(
            config=config, directory=directory, execute_command=execute_command
        )
        self._ssh_host = config["ssh_host"]
        self._ssh_username = config["ssh_username"]
        self._ssh_known_hosts = os.path.abspath(
            os.path.expanduser(config["known_hosts"])
        )
        self._ssh_key = os.path.abspath(os.path.expanduser(config["ssh_key"]))
        self._ssh_remote_config_dir = config["ssh_remote_config_dir"]
        self._ssh_remote_path = config["ssh_remote_path"]
        self._ssh_local_path = os.path.abspath(
            os.path.expanduser(config["ssh_local_path"])
        )
        if "ssh_delete_file_on_remote" in config.keys():
            self._ssh_delete_file_on_remote = config["ssh_delete_file_on_remote"]
        else:
            self._ssh_delete_file_on_remote = True
        if "ssh_port" in config.keys():
            self._ssh_port = config["ssh_port"]
        else:
            self._ssh_port = 22
        if "ssh_continous_connection" in config.keys():
            self._ssh_continous_connection = config["ssh_continous_connection"]
        else:
            self._ssh_continous_connection = False
        self._ssh_connection = None
        self._remote_flag = True

    def convert_path_to_remote(self, path):
        working_directory = os.path.abspath(os.path.expanduser(path))
        return self._get_remote_working_dir(working_directory=working_directory)

    def submit_job(
        self,
        queue=None,
        job_name=None,
        working_directory=None,
        cores=None,
        memory_max=None,
        run_time_max=None,
        dependency_list=None,
        command=None,
    ):
        if dependency_list is not None:
            raise NotImplementedError(
                "Submitting jobs with dependencies to a remote cluster is not yet supported."
            )
        self._transfer_data_to_remote(working_directory=working_directory)
        output = self._execute_remote_command(command=command)
        return int(output.split()[-1])

    def enable_reservation(self, process_id):
        """

        Args:
            process_id (int):

        Returns:
            str:
        """
        return self._execute_remote_command(
            command=self._reservation_command(job_id=process_id)
        )

    def delete_job(self, process_id):
        """

        Args:
            process_id (int):

        Returns:
            str:
        """
        return self._execute_remote_command(
            command=self._delete_command(job_id=process_id)
        )

    def get_queue_status(self, user=None):
        """

        Args:
            user (str):

        Returns:
            pandas.DataFrame:
        """
        df = pandas.DataFrame(
            json.loads(
                self._execute_remote_command(command=self._get_queue_status_command())
            )
        )
        if user is None:
            return df
        else:
            return df[df["user"] == user]

    def get_job_from_remote(self, working_directory):
        """
        Get the results of the calculation - this is necessary when the calculation was executed on a remote host.
        """
        working_directory = os.path.abspath(os.path.expanduser(working_directory))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        remote_dict = json.loads(
            self._execute_remote_command(
                command="python -m pysqa --list --working_directory "
                + remote_working_directory
            )
        )
        for d in remote_dict["dirs"]:
            local_dir = self._get_file_transfer(
                file=d, local_dir=remote_working_directory, remote_dir=working_directory
            )
            os.makedirs(local_dir, exist_ok=True)
        file_dict = {}
        for f in remote_dict["files"]:
            local_file = self._get_file_transfer(
                file=f, local_dir=remote_working_directory, remote_dir=working_directory
            )
            file_dict[local_file] = f
        self._transfer_files(file_dict=file_dict, sftp=None, transfer_back=True)
        if self._ssh_delete_file_on_remote:
            self._execute_remote_command(command="rm -r " + remote_working_directory)

    def transfer_file(self, file, transfer_back=False):
        working_directory = os.path.abspath(os.path.expanduser(file))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        self._create_remote_dir(directory=os.path.dirname(remote_working_directory))
        self._transfer_files(
            file_dict={working_directory: remote_working_directory},
            sftp=None,
            transfer_back=transfer_back,
        )
        if self._ssh_delete_file_on_remote and transfer_back:
            self._execute_remote_command(command="rm " + remote_working_directory)

    def __del__(self):
        if self._ssh_connection is not None:
            self._ssh_connection.close()

    def _check_ssh_connection(self):
        if self._ssh_connection is None:
            self._ssh_connection = self._open_ssh_connection()

    def _transfer_files(self, file_dict, sftp=None, transfer_back=False):
        if sftp is None:
            if self._ssh_continous_connection:
                self._check_ssh_connection()
                ssh = self._ssh_connection
            else:
                ssh = self._open_ssh_connection()
            sftp_client = ssh.open_sftp()
        else:
            sftp_client = sftp
        for file_src, file_dst in tqdm(file_dict.items()):
            if transfer_back:
                try:
                    sftp_client.get(file_dst, file_src)
                except FileNotFoundError:
                    pass
            else:
                sftp_client.put(file_src, file_dst)
        if sftp is None:
            sftp_client.close()

    def _open_ssh_connection(self):
        ssh = paramiko.SSHClient()
        ssh.load_host_keys(self._ssh_known_hosts)
        ssh.connect(
            hostname=self._ssh_host,
            port=self._ssh_port,
            username=self._ssh_username,
            key_filename=self._ssh_key,
        )
        return ssh

    def _remote_command(self):
        return "python -m pysqa --config_directory " + self._ssh_remote_config_dir + " "

    def _get_queue_status_command(self):
        return self._remote_command() + "--status"

    def _submit_command(
        self,
        queue=None,
        job_name=None,
        working_directory=None,
        cores=None,
        memory_max=None,
        run_time_max=None,
        command_str=None,
    ):
        command = self._remote_command() + "--submit "
        if queue is not None:
            command += "--queue " + queue + " "
        if job_name is not None:
            command += "--job_name " + job_name + " "
        if working_directory is not None:
            command += "--working_directory " + working_directory + " "
        if cores is not None:
            command += "--cores " + cores + " "
        if memory_max is not None:
            command += "--memory " + memory_max + " "
        if run_time_max is not None:
            command += "--run_time " + run_time_max + " "
        if command_str is not None:
            command += '--command "' + command_str + '" '
        return command

    def _delete_command(self, job_id):
        return self._remote_command() + "--delete --id " + str(job_id)

    def _reservation_command(self, job_id):
        return self._remote_command() + "--reservation --id " + str(job_id)

    def _execute_remote_command(self, command):
        if self._ssh_continous_connection:
            self._check_ssh_connection()
            ssh = self._ssh_connection
        else:
            ssh = self._open_ssh_connection()
        stdin, stdout, stderr = ssh.exec_command(command)
        warnings.warn(stderr.read().decode())
        output = stdout.read().decode()
        if not self._ssh_continous_connection:
            ssh.close()
        return output

    def _get_remote_working_dir(self, working_directory):
        return os.path.join(
            self._ssh_remote_path,
            os.path.relpath(working_directory, self._ssh_local_path),
        )

    def _create_remote_dir(self, directory):
        if isinstance(directory, str):
            self._execute_remote_command(command="mkdir -p " + directory)
        elif isinstance(directory, list):
            command = "mkdir -p "
            for d in directory:
                command += d + " "
            self._execute_remote_command(command=command)
        else:
            raise TypeError()

    def _transfer_data_to_remote(self, working_directory):
        working_directory = os.path.abspath(os.path.expanduser(working_directory))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        file_dict = {}
        new_dir_list = []
        for p, folder, files in os.walk(working_directory):
            new_dir_list.append(
                self._get_file_transfer(
                    file=p,
                    local_dir=working_directory,
                    remote_dir=remote_working_directory,
                )
            )
            for f in files:
                file_path = os.path.join(p, f)
                file_dict[file_path] = self._get_file_transfer(
                    file=file_path,
                    local_dir=working_directory,
                    remote_dir=remote_working_directory,
                )
        self._create_remote_dir(directory=new_dir_list)
        self._transfer_files(file_dict=file_dict, sftp=None, transfer_back=False)

    def _get_user(self):
        """

        Returns:
            str:
        """
        return self._ssh_username

    @staticmethod
    def _get_file_transfer(file, local_dir, remote_dir):
        return os.path.abspath(
            os.path.join(remote_dir, os.path.relpath(file, local_dir))
        )



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import pandas
import unittest
from pysqa import QueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"

df_queue_status = pandas.DataFrame(
    {
        "cluster": ["Mycluster", "Mycluster", "Mycluster", "Mycluster", "Mycluster"],
        "jobid": [5322019, 5322016, 5322017, 5322018, 5322013],
        "user": ["janj", "janj", "janj", "janj", "janj"],
        "jobname": ["pi_19576488", "pi_19576485", "pi_19576486", "pi_19576487", "pi_19576482"],
        "status": ["r", "r", "r", "r", "r"],
    }
)

class TestGentQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.gent = QueueAdapter(directory=os.path.join(cls.path, "config/gent"))

    def test_config(self):
        self.assertEqual(self.gent.config["queue_type"], "GENT")
        self.assertEqual(self.gent.config["queue_primary"], "slurm")

    def test_list_clusters(self):
        self.assertEqual(self.gent.list_clusters(), ['default'])

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.gent.ssh_delete_file_on_remote, True)

    def test_interfaces(self):
        self.assertEqual(
            self.gent._adapter._commands.get_queue_status_command,
            ["squeue", "--format", "%A|%u|%t|%.15j|%Z", "--noheader"],
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("gent"):
            self.assertEqual(
                self.gent._adapter._list_command_to_be_executed(None, "here"),
                ["sbatch", "--parsable", "here"],
            )
        with self.subTest("gent with dependency"):
            self.assertRaises(
                NotImplementedError,
                self.gent._adapter._list_command_to_be_executed,
                [],
                "here",
            )

    def test_get_job_id_from_output(self):
        self.assertEqual(self.gent._adapter._commands.get_job_id_from_output("123;MyQueue"), 123)

    def test_get_queue_from_output(self):
        self.assertEqual(self.gent._adapter._commands.get_queue_from_output("123;MyQueue"), "MyQueue")

    def test_convert_queue_status_slurm(self):
        with open(os.path.join(self.path, "config/gent", "gent_output"), "r") as f:
            content = f.read()
        self.assertTrue(
            df_queue_status.equals(
                self.gent._adapter._commands.convert_queue_status(
                    queue_status_output=content
                )
            )
        )

    def test_switch_cluster_command(self):
        self.assertEqual(
            self.gent._adapter._switch_cluster_command(cluster_module="module1"),
            ['module', '--quiet', 'swap', 'cluster/module1;']
        )

    def test_resolve_queue_id(self):
        self.assertEqual(
            self.gent._adapter._resolve_queue_id(process_id=20120, cluster_dict={0: "cluster1"}),
            ("cluster1", 2012)
        )

    def test_submit_job_no_output(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            pass

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertIsNone(gent_tmp.submit_job(
            queue="slurm",
            job_name="test",
            working_directory=".",
            command="echo hello"
        ))
        os.remove("run_queue.sh")

    def test_submit_job_with_output(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            return "123;cluster0"

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertEqual(gent_tmp.submit_job(
            queue="slurm",
            job_name="test",
            working_directory=".",
            command="echo hello"
        ), 1230)
        os.remove("run_queue.sh")

    def test_delete_job_no_output(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            pass

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertIsNone(gent_tmp.delete_job(process_id=1))

    def test_delete_job_with_output(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            return 0, 1

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertEqual(gent_tmp.delete_job(process_id=1), 0)

    def test_get_queue_status(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            with open(os.path.join(self.path, "config", "gent", "gent_output")) as f:
                return f.read()

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertTrue(pandas.concat([df_queue_status] * 3).reset_index(drop=True).equals(
            gent_tmp.get_queue_status())
        )

    def test_get_queue_status_user(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            with open(os.path.join(self.path, "config", "gent", "gent_output")) as f:
                return f.read()

        gent_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/gent"),
            execute_command=execute_command
        )
        self.assertTrue(pandas.concat([df_queue_status] * 3).reset_index(drop=True).equals(
            gent_tmp.get_queue_status(user="janj"))
        )



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import pandas
import unittest
import getpass
from pysqa import QueueAdapter

try:
    import defusedxml.ElementTree as ETree
    skip_sge_test = False
except ImportError:
    skip_sge_test = True

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


@unittest.skipIf(skip_sge_test, "defusedxml is not installed, so the sun grid engine (SGE) tests are skipped.")
class TestSGEQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.sge = QueueAdapter(directory=os.path.join(cls.path, "config/sge"))

    def test_config(self):
        self.assertEqual(self.sge.config["queue_type"], "SGE")
        self.assertEqual(self.sge.config["queue_primary"], "impi_hydra_small")

    def test_list_clusters(self):
        self.assertEqual(self.sge.list_clusters(), ['default'])

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.sge.ssh_delete_file_on_remote, True)

    def test_value_in_range(self):
        self.assertEqual(
            None,
            self.sge._adapter._value_in_range(
                value=None, value_min=None, value_max=None
            ),
        )
        self.assertEqual(
            1,
            self.sge._adapter._value_in_range(value=None, value_min=1, value_max=None),
        )
        self.assertEqual(
            1,
            self.sge._adapter._value_in_range(value=None, value_min=None, value_max=1),
        )
        self.assertEqual(
            1,
            self.sge._adapter._value_in_range(value=1, value_min=None, value_max=None),
        )
        self.assertEqual(
            1, self.sge._adapter._value_in_range(value=0, value_min=1, value_max=None)
        )
        self.assertEqual(
            1, self.sge._adapter._value_in_range(value=2, value_min=None, value_max=1)
        )
        self.assertEqual(
            1, self.sge._adapter._value_in_range(value=1, value_min=0, value_max=2)
        )

    def test_job_submission_template(self):
        self.assertRaises(
            ValueError, self.sge._adapter._job_submission_template, command=None
        )
        self.assertRaises(
            TypeError, self.sge._adapter._job_submission_template, command=1
        )
        template = (
            "#!/bin/bash\n#$ -N job.py\n#$ -wd .\n#$ -pe impi_hydra_small 1\n#$ -l h_rt=604800\n"
            "#$ -o time.out\n#$ -e error.out\n\npython test.py"
        )
        self.assertEqual(
            self.sge._adapter._job_submission_template(command="python test.py"),
            template,
        )
        self.assertRaises(
            ValueError,
            self.sge._adapter._job_submission_template,
            command="python test.py",
            queue="notavailable",
        )

    def test_interfaces(self):
        self.assertEqual(
            self.sge._adapter._commands.submit_job_command, ["qsub", "-terse"]
        )
        self.assertEqual(self.sge._adapter._commands.delete_job_command, ["qdel"])
        self.assertEqual(
            self.sge._adapter._commands.enable_reservation_command,
            ["qalter", "-R", "y"],
        )
        self.assertEqual(
            self.sge._adapter._commands.get_queue_status_command, ["qstat", "-xml"]
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("sge"):
            self.assertEqual(
                self.sge._adapter._list_command_to_be_executed(None, "here"),
                ["qsub", "-terse", "here"],
            )
        with self.subTest("sge with dependency"):
            self.assertRaises(
                NotImplementedError,
                self.sge._adapter._list_command_to_be_executed,
                [],
                "here",
            )

    def test_convert_queue_status_sge(self):
        with open(os.path.join(self.path, "config/sge", "qstat.xml"), "r") as f:
            content = f.read()
        df_running = pandas.DataFrame(
            {
                "jobid": ["2836045"],
                "user": ["friko"],
                "jobname": ["vasp.5.3.5"],
                "status": ["running"],
            }
        )
        df_pending = pandas.DataFrame(
            {
                "jobid": ["2836046", "2967274"],
                "user": ["friko", "janj"],
                "jobname": ["vasp.5.3.5", "hello.py"],
                "status": ["pending", "error"],
            }
        )
        df_merge = pandas.concat([df_running, df_pending], sort=True)
        df = pandas.DataFrame(
            {
                "jobid": pandas.to_numeric(df_merge.jobid),
                "user": df_merge.user,
                "jobname": df_merge.jobname,
                "status": df_merge.status,
                "working_directory": [""] * len(df_merge),
            }
        )
        self.assertTrue(
            df.equals(
                self.sge._adapter._commands.convert_queue_status(
                    queue_status_output=content
                )
            )
        )

    def test_queue_list(self):
        self.assertEqual(
            sorted(self.sge.queue_list),
            ["impi_hy", "impi_hydra", "impi_hydra_cmfe", "impi_hydra_small"],
        )

    def test_queues(self):
        self.assertEqual(self.sge.queues.impi_hydra, "impi_hydra")
        self.assertEqual(
            sorted(dir(self.sge.queues)),
            ["impi_hy", "impi_hydra", "impi_hydra_cmfe", "impi_hydra_small"],
        )
        with self.assertRaises(AttributeError):
            _ = self.sge.queues.notavailable

    def test_get_user(self):
        self.assertEqual(self.sge._adapter._get_user(), getpass.getuser())

    def test_check_queue_parameters(self):
        self.assertEqual(
            (1, 604800, None), self.sge.check_queue_parameters(queue="impi_hydra_small")
        )



import os
import unittest
from pysqa.utils.execute import execute_command


class TestExecuteCommand(unittest.TestCase):
    def test_commands_as_lst(self):
        output = execute_command(
            commands=["echo", "hello"],
            working_directory=".",
            split_output=True,
            shell=False,
            error_filename="pysqa.err"
        )
        self.assertEqual(output, ['hello', ''])

    def test_commands_as_lst_no_split(self):
        output = execute_command(
            commands=["echo", "hello"],
            working_directory=".",
            split_output=False,
            shell=False,
            error_filename="pysqa.err"
        )
        self.assertEqual(output, 'hello\n')

    def test_commands_as_lst_shell_true(self):
        output = execute_command(
            commands=["echo", "hello"],
            working_directory=".",
            split_output=True,
            shell=True,
            error_filename="pysqa.err"
        )
        self.assertEqual(output, ['hello', ''])

    def test_commands_as_str(self):
        output = execute_command(
            commands="echo hello",
            working_directory=".",
            split_output=True,
            shell=False,
            error_filename="pysqa.err"
        )
        self.assertEqual(output, ['hello', ''])

    def test_commands_fails(self):
        output = execute_command(
            commands="exit 1",
            working_directory=".",
            split_output=True,
            shell=False,
            error_filename="pysqa_fails.err"
        )
        self.assertIsNone(output)
        with open("pysqa_fails.err") as f:
            error = f.readlines()
        self.assertEqual(error, ["\n"])
        os.remove("pysqa_fails.err")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import pandas

from pysqa.wrapper.generic import SchedulerCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Feb 9, 2019"


class SlurmCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["sbatch", "--parsable"]

    @property
    def delete_job_command(self):
        return ["scancel"]

    @property
    def get_queue_status_command(self):
        return ["squeue", "--format", "%A|%u|%t|%.15j|%Z", "--noheader"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        return int(queue_submit_output.splitlines()[-1].rstrip().lstrip().split()[-1])

    @staticmethod
    def convert_queue_status(queue_status_output):
        line_split_lst = [line.split("|") for line in queue_status_output.splitlines()]
        if len(line_split_lst) != 0:
            job_id_lst, user_lst, status_lst, job_name_lst, working_directory_lst = zip(
                *[
                    (int(jobid), user, status.lower(), jobname, working_directory)
                    for jobid, user, status, jobname, working_directory in line_split_lst
                ]
            )
        else:
            job_id_lst, user_lst, status_lst, job_name_lst, working_directory_lst = (
                [],
                [],
                [],
                [],
                [],
            )
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
                "working_directory": working_directory_lst,
            }
        )
        df.loc[df.status == "r", "status"] = "running"
        df.loc[df.status == "pd", "status"] = "pending"
        return df

    @staticmethod
    def dependencies(dependency_list) -> list:
        if dependency_list is not None:
            return ["--dependency=afterok:" + ",".join(dependency_list)]
        else:
            return []



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import re

import pandas as pd

from pysqa.wrapper.generic import SchedulerCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Feb 9, 2019"


class TorqueCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["qsub"]

    @property
    def delete_job_command(self):
        return ["qdel"]

    @property
    def get_queue_status_command(self):
        return ["qstat", "-f"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        # strip last line from output, leading and trailing whitespaces,
        # and separates the queue id from the stuff after "."
        # Adjust if your system doesn't have output like below!
        # e.g. qsub run_queue.sh -> "12347673.gadi-pbs", the below returns 12347673
        # It must return an integer for it to not raise an exception later.
        return int(
            queue_submit_output.splitlines()[-1].rstrip().lstrip().split(sep=".")[0]
        )

    @staticmethod
    def convert_queue_status(queue_status_output):
        # # Run the qstat -f command and capture its output
        # output = subprocess.check_output(["qstat", "-f"]) -> output is the
        # queue_status_output that goes into this function

        # Split the output into lines
        lines = queue_status_output.split("\n")  # .decode().split("\n")

        # concatenate all lines into a single string
        input_string = "".join(lines)
        # remove all whitespaces
        input_string = "".join(input_string.split())

        # Extract the job ID, user, job name, status, and working directory for each running job
        regex_pattern_job_id = r"JobId:(.*?)Job_Name"
        job_id_lst = re.findall(regex_pattern_job_id, input_string)
        job_id_lst = [int(job_id_str.split(sep=".")[0]) for job_id_str in job_id_lst]

        regex_pattern_user = r"Job_Owner=(.*?)job_state"
        user_lst = re.findall(regex_pattern_user, input_string)
        user_lst = [user_str.split("@")[0] for user_str in user_lst]

        regex_pattern_job_name = r"Job_Name=(.*?)Job_Owner"
        job_name_lst = re.findall(regex_pattern_job_name, input_string)

        regex_pattern_status = r"job_state=(.*?)queue"
        status_lst = re.findall(regex_pattern_status, input_string)

        regex_pattern_working_directory = r"PBS_O_WORKDIR=(.*?),PBS"
        working_directory_lst = re.findall(
            regex_pattern_working_directory, input_string
        )

        df = pd.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
                "working_directory": working_directory_lst,
            }
        )

        df["status"] = df["status"].apply(
            lambda x: "running" if x == "R" else "pending"
        )

        return df



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest
from pysqa import QueueAdapter

try:
    import paramiko
    from tqdm import tqdm
    skip_multi_test = False
except ImportError:
    skip_multi_test = True


__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


@unittest.skipIf(skip_multi_test, "Either paramiko or tqdm are not installed, so the multi queue adapter tests are skipped.")
class TestMultiQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.multi = QueueAdapter(directory=os.path.join(cls.path, "config/multicluster"))

    def test_config(self):
        self.assertEqual(self.multi.config["queue_type"], "SLURM")
        self.assertEqual(self.multi.config["queue_primary"], "slurm")

    def test_list_clusters(self):
        self.assertEqual(self.multi.list_clusters(), ['local_slurm', 'remote_slurm'])



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import pandas

from pysqa.wrapper.generic import SchedulerCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Feb 9, 2019"


class LsfCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["bsub"]

    @property
    def delete_job_command(self):
        return ["bkill"]

    @property
    def get_queue_status_command(self):
        return ["bjobs"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        return int(queue_submit_output.split("<")[1].split(">")[0])

    @staticmethod
    def convert_queue_status(queue_status_output):
        job_id_lst, user_lst, status_lst, job_name_lst = [], [], [], []
        line_split_lst = queue_status_output.split("\n")
        if len(line_split_lst) > 1:
            for l in line_split_lst[1:]:
                line_segments = l.split()
                if len(line_segments) > 1:
                    job_id_lst.append(int(line_segments[0]))
                    user_lst.append(line_segments[1])
                    status_lst.append(line_segments[2])
                    job_name_lst.append(line_segments[6])
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )
        df.loc[df.status == "RUN", "status"] = "running"
        df.loc[df.status == "PEND", "status"] = "pending"
        return df



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import pandas

from pysqa.wrapper.slurm import SlurmCommands

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Feb 9, 2019"


class GentCommands(SlurmCommands):
    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        return int(queue_submit_output.splitlines()[-1].rstrip().lstrip().split(";")[0])

    @staticmethod
    def get_queue_from_output(queue_submit_output):
        return str(queue_submit_output.splitlines()[-1].rstrip().lstrip().split(";")[1])

    @staticmethod
    def convert_queue_status(queue_status_output):
        qstat = queue_status_output.splitlines()
        queue = qstat[0].split(":")[1].strip()
        if len(qstat) <= 1:  # first row contains cluster name, check if there are jobs
            return None

        line_split_lst = [line.split("|") for line in qstat[1:]]
        job_id_lst, user_lst, status_lst, job_name_lst, queue_lst = zip(
            *[
                (int(jobid), user, status.lower(), jobname, queue)
                for jobid, user, status, jobname in line_split_lst
            ]
        )
        return pandas.DataFrame(
            {
                "cluster": queue_lst,
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )

    @staticmethod
    def dependencies(dependency_list) -> list:
        if dependency_list is not None:
            raise NotImplementedError()
        else:
            return []



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest
import pandas
from pysqa import QueueAdapter


try:
    import flux
    skip_flux_test = False
except ImportError:
    skip_flux_test = True


@unittest.skipIf(skip_flux_test, "Flux is not installed, so the flux tests are skipped.")
class TestFluxQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.flux = QueueAdapter(directory=os.path.join(cls.path, "config/flux"))

    def test_config(self):
        self.assertEqual(self.flux.config["queue_type"], "FLUX")
        self.assertEqual(self.flux.config["queue_primary"], "flux")

    def test_list_clusters(self):
        self.assertEqual(self.flux.list_clusters(), ['default'])

    def test_remote_flag(self):
        self.assertFalse(self.flux._adapter.remote_flag)

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.flux.ssh_delete_file_on_remote, True)

    def test_interfaces(self):
        self.assertEqual(
            self.flux._adapter._commands.submit_job_command, ["flux", "batch"]
        )
        self.assertEqual(self.flux._adapter._commands.delete_job_command, ["flux", "cancel"])
        self.assertEqual(
            self.flux._adapter._commands.get_queue_status_command,
            ["flux", "jobs", "-a", "--no-header"],
        )

    def test_convert_queue_status_slurm(self):
        with open(os.path.join(self.path, "config/flux", "flux_jobs"), "r") as f:
            content = f.read()
        df = pandas.DataFrame({
            "jobid": [1125147213824, 1109007532032, 1092532305920],
            "user": ["dahn", "dahn", "dahn"],
            "jobname": ["sleep_batc", "sleep_batc", "sleep_batc"],
            "status": ["running", "running", "running"]
        })
        self.assertTrue(df.equals(self.flux._adapter._commands.convert_queue_status(
            queue_status_output=content
        )))

    def test_submit_job(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            return "ƒWZEQa8X\n"

        flux_tmp = QueueAdapter(
            directory=os.path.join(self.path, "config/flux"),
            execute_command=execute_command
        )
        self.assertEqual(flux_tmp.submit_job(
            queue="flux",
            job_name="test",
            working_directory=".",
            cores=4,
            command="echo hello"
        ), 1125147213824)
        with open("run_queue.sh") as f:
            output = f.read()
        content = """\
#!/bin/bash
# flux:--job-name=test
# flux: --env=CORES=4
# flux: --output=time.out
# flux: --error=error.out
# flux: -n 4
# flux: -t 2880

echo hello"""
        self.assertEqual(content, output)
        os.remove("run_queue.sh")



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest

import pandas

from pysqa import QueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


class TestLsfQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.lsf = QueueAdapter(directory=os.path.join(cls.path, "config/lsf"))

    def test_config(self):
        self.assertEqual(self.lsf.config["queue_type"], "LSF")
        self.assertEqual(self.lsf.config["queue_primary"], "lsf")

    def test_list_clusters(self):
        self.assertEqual(self.lsf.list_clusters(), ['default'])

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.lsf.ssh_delete_file_on_remote, True)

    def test_job_submission_template(self):
        template = (
            "#!/bin/bash\n#BSUB -q queue\n#BSUB -J job.py\n#BSUB -o time.out\n#BSUB -n 10\n#BSUB -cwd .\n"
            "#BSUB -e error.out\n#BSUB -W 259200\n\npython test.py"
        )
        self.assertEqual(
            self.lsf._adapter._job_submission_template(command="python test.py"),
            template,
        )

    def test_interfaces(self):
        self.assertEqual(
            self.lsf._adapter._commands.submit_job_command, ["bsub"]
        )
        self.assertEqual(self.lsf._adapter._commands.delete_job_command, ["bkill"])
        self.assertEqual(
            self.lsf._adapter._commands.get_queue_status_command, ["bjobs"]
        )

    def test__list_command_to_be_executed(self):
        with self.subTest("lsf"):
            self.assertEqual(
                self.lsf._adapter._list_command_to_be_executed(None, "here"),
                ["bsub", "here"],
            )
        with self.subTest("lsf with dependency"):
            self.assertRaises(
                NotImplementedError,
                self.lsf._adapter._list_command_to_be_executed,
                [],
                "here",
            )

    def test_convert_queue_status_sge(self):
        with open(os.path.join(self.path, "config/lsf", "bjobs_output"), "r") as f:
            content = f.read()
        df = pandas.DataFrame({
            "jobid": [5136563, 5136570, 5136571],
            "user": ["testuse"] * 3,
            "jobname": ["pi_None"] * 3,
            "status": ["running"] * 3
        })
        self.assertTrue(
            df.equals(
                self.lsf._adapter._commands.convert_queue_status(
                    queue_status_output=content
                )
            )
        )



import os
import sys

from pympipool.mpi import PyMPIExecutor
from pysqa.executor.helper import (
    read_from_file,
    deserialize,
    write_to_file,
    serialize_result,
)


def execute_files_from_list(tasks_in_progress_dict, cache_directory, executor):
    file_lst = os.listdir(cache_directory)
    for file_name_in in file_lst:
        key = file_name_in.split(".in.pl")[0]
        file_name_out = key + ".out.pl"
        if (
            file_name_in.endswith(".in.pl")
            and file_name_out not in file_lst
            and key not in tasks_in_progress_dict.keys()
        ):
            funct_dict = read_from_file(
                file_name=os.path.join(cache_directory, file_name_in)
            )
            apply_dict = deserialize(funct_dict=funct_dict)
            for k, v in apply_dict.items():
                tasks_in_progress_dict[k] = executor.submit(
                    v["fn"], *v["args"], **v["kwargs"]
                )
    for k, v in tasks_in_progress_dict.items():
        if v.done():
            write_to_file(
                funct_dict=serialize_result(result_dict={k: v.result()}),
                state="out",
                cache_directory=cache_directory,
            )


def execute_tasks(cores, cache_directory):
    tasks_in_progress_dict = {}
    with PyMPIExecutor(
        max_workers=cores,
        cores_per_worker=1,
        threads_per_core=1,
        gpus_per_worker=0,
        oversubscribe=False,
        init_function=None,
        cwd=cache_directory,
        sleep_interval=0.1,
        enable_slurm_backend=False,
    ) as exe:
        while True:
            execute_files_from_list(
                tasks_in_progress_dict=tasks_in_progress_dict,
                cache_directory=cache_directory,
                executor=exe,
            )


def command_line(arguments_lst=None):
    if arguments_lst is None:
        arguments_lst = sys.argv[1:]
    cores_arg = arguments_lst[arguments_lst.index("--cores") + 1]
    path_arg = arguments_lst[arguments_lst.index("--path") + 1]
    execute_tasks(cores=cores_arg, cache_directory=path_arg)



# coding: utf-8
# Copyright (c) Jan Janssen

import pandas

from pysqa.utils.basic import BasisQueueAdapter
from pysqa.utils.execute import execute_command


class ModularQueueAdapter(BasisQueueAdapter):
    def __init__(self, config, directory="~/.queues", execute_command=execute_command):
        super(ModularQueueAdapter, self).__init__(
            config=config, directory=directory, execute_command=execute_command
        )
        self._queue_to_cluster_dict = {
            k: v["cluster"] for k, v in self._config["queues"].items()
        }
        for v in self._queue_to_cluster_dict.values():
            if v not in self._config["cluster"]:
                raise ValueError(
                    "The cluster "
                    + v
                    + " was not found in the list of clusters "
                    + str(list(self._config["cluster"].keys()))
                )

    def submit_job(
        self,
        queue=None,
        job_name=None,
        working_directory=None,
        cores=None,
        memory_max=None,
        run_time_max=None,
        dependency_list=None,
        command=None,
        **kwargs,
    ):
        """

        Args:
            queue (str/None):
            job_name (str/None):
            working_directory (str/None):
            cores (int/None):
            memory_max (int/None):
            run_time_max (int/None):
            dependency_list (list/None):
            command (str/None):

        Returns:
            int:
        """
        working_directory, queue_script_path = self._write_queue_script(
            queue=queue,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            command=command,
            **kwargs,
        )
        cluster_module = self._queue_to_cluster_dict[queue]
        commands = self._switch_cluster_command(
            cluster_module=cluster_module
        ) + self._list_command_to_be_executed(dependency_list, queue_script_path)
        out = self._execute_command(
            commands=commands,
            working_directory=working_directory,
            split_output=False,
            shell=True,
        )
        if out is not None:
            cluster_queue_id = self._commands.get_job_id_from_output(out)
            cluster_queue_id *= 10
            cluster_queue_id += self._config["cluster"].index(cluster_module)
            return cluster_queue_id
        else:
            return None

    def enable_reservation(self, process_id):
        """

        Args:
            process_id (int):

        Returns:
            str:
        """
        cluster_module, cluster_queue_id = self._resolve_queue_id(
            process_id=process_id, cluster_dict=self._config["cluster"]
        )
        cluster_commands = self._switch_cluster_command(cluster_module=cluster_module)
        commands = (
            cluster_commands
            + self._commands.enable_reservation_command
            + [str(cluster_queue_id)]
        )
        out = self._execute_command(commands=commands, split_output=True, shell=True)
        if out is not None:
            return out[0]
        else:
            return None

    def delete_job(self, process_id):
        """

        Args:
            process_id (int):

        Returns:
            str:
        """
        cluster_module, cluster_queue_id = self._resolve_queue_id(
            process_id=process_id, cluster_dict=self._config["cluster"]
        )
        cluster_commands = self._switch_cluster_command(cluster_module=cluster_module)
        commands = (
            cluster_commands
            + self._commands.delete_job_command
            + [str(cluster_queue_id)]
        )
        out = self._execute_command(commands=commands, split_output=True, shell=True)
        if out is not None:
            return out[0]
        else:
            return None

    def get_queue_status(self, user=None):
        """

        Args:
            user (str):

        Returns:
            pandas.DataFrame:
        """
        df_lst = []
        for cluster_module in self._config["cluster"]:
            cluster_commands = self._switch_cluster_command(
                cluster_module=cluster_module
            )
            out = self._execute_command(
                commands=cluster_commands + self._commands.get_queue_status_command,
                split_output=False,
                shell=True,
            )
            df = self._commands.convert_queue_status(queue_status_output=out)
            df_lst.append(df)
        df = pandas.concat(df_lst, axis=0, sort=False).reset_index(drop=True)
        if user is None:
            return df
        else:
            return df[df["user"] == user]

    @staticmethod
    def _resolve_queue_id(process_id, cluster_dict):
        cluster_queue_id = int(process_id / 10)
        cluster_module = cluster_dict[process_id - cluster_queue_id * 10]
        return cluster_module, cluster_queue_id

    @staticmethod
    def _switch_cluster_command(cluster_module):
        return ["module", "--quiet", "swap", "cluster/{};".format(cluster_module)]



import getopt
import json
import os
import sys

from pysqa.queueadapter import QueueAdapter
from pysqa.utils.execute import execute_command


def command_line(arguments_lst=None, execute_command=execute_command):
    """
    Parse the command line arguments.

    Args:
        argv: Command line arguments
        execute_command: function to comunicate with shell process

    """
    directory = "~/.queues"
    queue = None
    job_name = None
    working_directory = None
    cores = None
    memory_max = None
    run_time_max = None
    command = None
    job_id = None
    if arguments_lst is None:
        arguments_lst = sys.argv[1:]
    try:
        opts, args = getopt.getopt(
            arguments_lst,
            "f:pq:j:w:n:m:t:b:c:ri:dslh",
            [
                "config_directory=",
                "submit",
                "queue=",
                "job_name=",
                "working_directory=",
                "cores=",
                "memory=",
                "run_time=",
                "dependency=",
                "command=",
                "reservation",
                "id=",
                "delete",
                "status",
                "list",
                "help",
            ],
        )
    except getopt.GetoptError:
        print("python -m pysqa --help")
    else:
        mode_submit = False
        mode_delete = False
        mode_reservation = False
        mode_status = False
        mode_list = False
        dependency_list = None
        for opt, arg in opts:
            if opt in ("-f", "--config_directory"):
                directory = arg
            elif opt in ("-p", "--submit"):
                mode_submit = True
            elif opt in ("-q", "--queue"):
                queue = arg
            elif opt in ("-j", "--job_name"):
                job_name = arg
            elif opt in ("-w", "--working_directory"):
                working_directory = arg
            elif opt in ("-n", "--cores"):
                cores = int(arg)
            elif opt in ("-m", "--memory"):
                memory_max = arg
            elif opt in ("-t", "--run_time"):
                run_time_max = arg
            elif opt in ("-c", "--command"):
                command = arg
            elif opt in ("-r", "--reservation"):
                mode_reservation = arg
            elif opt in ("-i", "--id"):
                if arg != "":
                    job_id = int(arg)
            elif opt in ("-d", "--delete"):
                mode_delete = True
            elif opt in ("-s", "--status"):
                mode_status = True
            elif opt in ("-l", "--list"):
                mode_list = True
            elif opt in ("-b", "--dependency"):
                if dependency_list is None:
                    dependency_list = [arg]
                else:
                    dependency_list.append(arg)
        if mode_submit or mode_delete or mode_reservation or mode_status:
            qa = QueueAdapter(directory=directory, execute_command=execute_command)
            if mode_submit:
                print(
                    qa.submit_job(
                        queue=queue,
                        job_name=job_name,
                        working_directory=working_directory,
                        cores=cores,
                        memory_max=memory_max,
                        run_time_max=run_time_max,
                        dependency_list=dependency_list,
                        command=command,
                    )
                )
            elif mode_delete:
                print(qa.delete_job(process_id=job_id))
            elif mode_reservation:
                print(qa.enable_reservation(process_id=job_id))
            elif mode_status:
                print(json.dumps(qa.get_queue_status().to_dict(orient="list")))
        elif mode_list:
            working_directory = os.path.abspath(os.path.expanduser(working_directory))
            remote_dirs, remote_files = [], []
            for p, folder, files in os.walk(working_directory):
                remote_dirs.append(p)
                remote_files += [os.path.join(p, f) for f in files]
            print(
                json.dumps({"dirs": sorted(remote_dirs), "files": sorted(remote_files)})
            )
        else:
            print("python -m pysqa --help ... coming soon.")



import unittest
from pysqa.wrapper.generic import SchedulerCommands


class TmpSchedularCommands(SchedulerCommands):
    def delete_job_command(self):
        pass

    def get_queue_status_command(self):
        pass

    def submit_job_command(self):
        pass


class TestSchedulerCommands(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.commands = TmpSchedularCommands()

    def test_enable_reservation_command(self):
        with self.assertRaises(NotImplementedError):
            self.commands.enable_reservation_command()

    def test_get_job_id_from_output(self):
        with self.assertRaises(NotImplementedError):
            self.commands.get_job_id_from_output("test")

    def test_convert_queue_status(self):
        with self.assertRaises(NotImplementedError):
            self.commands.convert_queue_status("test")

    def test_no_delete_job_command(self):
        class NoDelteteSchedularCommands(SchedulerCommands):

            def get_queue_status_command(self):
                pass

            def submit_job_command(self):
                pass

        with self.assertRaises(TypeError):
            NoDelteteSchedularCommands()

    def test_get_queue_status_command(self):
        class NoQueueStatusSchedularCommands(SchedulerCommands):
            def delete_job_command(self):
                pass

            def submit_job_command(self):
                pass

        with self.assertRaises(TypeError):
            NoQueueStatusSchedularCommands()

    def test_submit_job_command(self):
        class NoSubmitSchedularCommands(SchedulerCommands):
            def delete_job_command(self):
                pass

            def get_queue_status_command(self):
                pass

        with self.assertRaises(TypeError):
            NoSubmitSchedularCommands()



import os
import subprocess


def execute_command(
    commands,
    working_directory=None,
    split_output=True,
    shell=False,
    error_filename="pysqa.err",
):
    """
    A wrapper around the subprocess.check_output function.

    Args:
        commands (list/str): These commands are executed on the command line
        working_directory (str/None): The directory where the command is executed
        split_output (bool): Boolean flag to split newlines in the output
        shell (bool): Additional switch to convert list of commands to one string
        error_filename (str): In case the execution fails the output is written to this file

    Returns:
        str/list: output of the shell command either as string or as a list of strings
    """
    if shell and isinstance(commands, list):
        commands = " ".join(commands)
    try:
        out = subprocess.check_output(
            commands,
            cwd=working_directory,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            shell=not isinstance(commands, list),
        )
    except subprocess.CalledProcessError as e:
        with open(os.path.join(working_directory, error_filename), "w") as f:
            print(e.stdout, file=f)
        out = None
    if out is not None and split_output:
        return out.split("\n")
    else:
        return out



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest
from jinja2.exceptions import TemplateSyntaxError
from pysqa import QueueAdapter
from pysqa.utils.basic import BasisQueueAdapter

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


class TestQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))

    def test_missing_config(self):
        with self.assertRaises(ValueError):
            QueueAdapter(directory=os.path.join(self.path, "config/error"))

    def test_bad_queue_template(self):
        with self.assertRaises(TemplateSyntaxError):
            QueueAdapter(directory=os.path.join(self.path, "config/bad_template"))


class TestBasisQueueAdapter(unittest.TestCase):
    def test_bad_queue_type(self):
        with self.assertRaises(ValueError):
            BasisQueueAdapter(config={"queue_type": "error", "queues": {}})

    def test_memory_string_comparison(self):
        self.assertEqual(BasisQueueAdapter._value_in_range(1023, value_min="1K"), "1K")
        self.assertEqual(BasisQueueAdapter._value_in_range(1035, value_min="1K"), 1035)
        self.assertEqual(BasisQueueAdapter._value_in_range(1035, value_max="1K"), "1K")
        self.assertEqual(
            BasisQueueAdapter._value_in_range("1035", value_min="1K"), "1035"
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range(
                "60000M", value_min="1K", value_max="50G"
            ),
            "50G",
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range("60000", value_min="1K", value_max="50G"),
            "50G",
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range(
                "60000M", value_min="1K", value_max="70G"
            ),
            "60000M",
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range(60000, value_min="1K", value_max="70G"),
            60000,
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range(
                90000 * 1024 ** 2, value_min="1K", value_max="70G"
            ),
            "70G",
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range("90000", value_min="1K", value_max="70G"),
            "70G",
        )
        self.assertEqual(
            BasisQueueAdapter._value_in_range(
                "60000M", value_min="60G", value_max="70G"
            ),
            "60G",
        )



import os
import io
import json
import unittest
import unittest.mock
from pysqa.cmd import command_line


class TestCMD(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_dir = os.path.abspath(os.path.dirname(__file__))

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def assert_stdout_command_line(self, cmd_args, execute_command, expected_output, mock_stdout):
        command_line(
            arguments_lst=cmd_args,
            execute_command=execute_command
        )
        self.assertEqual(mock_stdout.getvalue(), expected_output)

    def test_help(self):
        self.assert_stdout_command_line(
            ["--help"],
            None,
            "python -m pysqa --help ... coming soon.\n",
        )

    def test_wrong_option(self):
        self.assert_stdout_command_line(
            ["--error"],
            None,
            "python -m pysqa --help\n",
        )

    def test_submit(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            return "1\n"

        self.assert_stdout_command_line(
            [
                "--config_directory", os.path.join(self.test_dir, "config", "slurm"),
                "--submit",
                "--queue", "slurm",
                "--job_name", "test",
                "--working_directory", ".",
                "--cores", "2",
                "--memory", "1GB",
                "--run_time", "10",
                "--command", "echo hello"
            ],
            execute_command,
            "1\n",
        )
        with open("run_queue.sh") as f:
            output = f.readlines()
        content = [
            '#!/bin/bash\n',
            '#SBATCH --output=time.out\n',
            '#SBATCH --job-name=test\n',
            '#SBATCH --chdir=.\n',
            '#SBATCH --get-user-env=L\n',
            '#SBATCH --partition=slurm\n',
            '#SBATCH --time=4320\n',
            '#SBATCH --mem=1GBG\n',
            '#SBATCH --cpus-per-task=10\n',
            '\n',
            'echo hello'
        ]
        self.assertEqual(output, content)
        os.remove("run_queue.sh")

    def test_delete(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            return "Success\n"

        self.assert_stdout_command_line(
            [
                "--config_directory", os.path.join(self.test_dir, "config", "slurm"),
                "--delete",
                "--id", "1"
            ],
            execute_command,
            "S\n"
        )

    def test_status(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            with open(os.path.join(self.test_dir, "config", "slurm", "squeue_output")) as f:
                return f.read()

        self.assert_stdout_command_line(
            [
                "--config_directory", os.path.join(self.test_dir, "config", "slurm"),
                "--status"
            ],
            execute_command,
            json.dumps({
                "jobid": [5322019, 5322016, 5322017, 5322018, 5322013], "user": ["janj", "janj", "janj", "janj", "maxi"],
                "jobname": ["pi_19576488", "pi_19576485", "pi_19576486", "pi_19576487", "pi_19576482"],
                "status": ["running", "running", "running", "running", "running"],
                "working_directory": [
                    "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_1",
                    "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_2",
                    "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_3",
                    "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_4",
                    "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_5"
                ]
            }) +"\n"
        )

    def test_list(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            pass

        self.assert_stdout_command_line(
            [
                "--config_directory", os.path.join(self.test_dir, "config", "slurm"),
                "--list",
                "--working_directory", os.path.join(self.test_dir, "config", "slurm"),

            ],
            execute_command,
            json.dumps({
                "dirs": [os.path.join(self.test_dir, "config", "slurm")],
                "files": sorted([
                    os.path.join(self.test_dir, "config", "slurm", "squeue_output"),
                    os.path.join(self.test_dir, "config", "slurm", "slurm_extra.sh"),
                    os.path.join(self.test_dir, "config", "slurm", "slurm.sh"),
                    os.path.join(self.test_dir, "config", "slurm", "queue.yaml"),
                ])
            }) + "\n"
        )



# coding: utf-8
import pandas
from flux.job import JobID

from pysqa.wrapper.generic import SchedulerCommands


class FluxCommands(SchedulerCommands):
    @property
    def submit_job_command(self):
        return ["flux", "batch"]

    @property
    def delete_job_command(self):
        return ["flux", "cancel"]

    @property
    def get_queue_status_command(self):
        return ["flux", "jobs", "-a", "--no-header"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output):
        return int(
            JobID(queue_submit_output.splitlines()[-1].rstrip().lstrip().split()[-1])
        )

    @staticmethod
    def convert_queue_status(queue_status_output):
        line_split_lst = [line.split() for line in queue_status_output.splitlines()]
        job_id_lst, user_lst, job_name_lst, status_lst = [], [], [], []
        for entry in line_split_lst:
            job_id_lst.append(JobID(entry[0]))
            user_lst.append(entry[1])
            job_name_lst.append(entry[2])
            status_lst.append(entry[3])
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )
        df.loc[df.status == "R", "status"] = "running"
        df.loc[df.status == "S", "status"] = "pending"
        df.loc[df.status == "CD", "status"] = "finished"
        return df



# coding: utf-8
# Copyright (c) Jan Janssen

import os
import unittest
from pysqa import QueueAdapter

try:
    import paramiko
    from tqdm import tqdm
    skip_remote_test = False
except ImportError:
    skip_remote_test = True


__author__ = "Jan Janssen"
__copyright__ = "Copyright 2019, Jan Janssen"
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 9, 2019"


@unittest.skipIf(skip_remote_test, "Either paramiko or tqdm are not installed, so the remote queue adapter tests are skipped.")
class TestRemoteQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.remote = QueueAdapter(directory=os.path.join(cls.path, "config/remote"))
        cls.remote_alternative = QueueAdapter(directory=os.path.join(cls.path, "config/remote_alternative"))

    def test_config(self):
        self.assertEqual(self.remote.config["queue_type"], "REMOTE")
        self.assertEqual(self.remote.config["queue_primary"], "remote")

    def test_list_clusters(self):
        self.assertEqual(self.remote.list_clusters(), ['default'])

    def test_remote_flag(self):
        self.assertTrue(self.remote._adapter.remote_flag)

    def test_ssh_delete_file_on_remote(self):
        self.assertEqual(self.remote.ssh_delete_file_on_remote, False)

    def test_ssh_continous_connection(self):
        self.assertEqual(self.remote._adapter._ssh_continous_connection, True)
        self.assertEqual(self.remote_alternative._adapter._ssh_continous_connection, False)

    def test_submit_job_remote(self):
        with self.assertRaises(NotImplementedError):
            self.remote.submit_job(queue="remote", dependency_list=[])

    def test_submit_command(self):
        command = self.remote._adapter._submit_command(
            queue="remote",
            job_name="test",
            working_directory="/home/localuser/projects/test",
            cores=str(1),
            memory_max=str(1),
            run_time_max=str(1),
            command_str="/bin/true",
        )
        self.assertEqual(
            command,
            'python -m pysqa --config_directory /u/share/pysqa/resources/queues/ --submit --queue remote --job_name test --working_directory /home/localuser/projects/test --cores 1 --memory 1 --run_time 1 --command "/bin/true" '
        )

    def test_get_queue_status_command(self):
        command = self.remote._adapter._get_queue_status_command()
        self.assertEqual(command, "python -m pysqa --config_directory /u/share/pysqa/resources/queues/ --status")

    def test_convert_path_to_remote(self):
        self.assertEqual(self.remote.convert_path_to_remote("/home/localuser/projects/test"), "/u/hpcuser/remote/test")

    def test_delete_command(self):
        self.assertEqual(
            "python -m pysqa --config_directory /u/share/pysqa/resources/queues/ --delete --id 123",
            self.remote._adapter._delete_command(job_id=123)
        )

    def test_reservation_command(self):
        self.assertEqual(
            "python -m pysqa --config_directory /u/share/pysqa/resources/queues/ --reservation --id 123",
            self.remote._adapter._reservation_command(job_id=123)
        )

    def test_get_ssh_user(self):
        self.assertEqual(self.remote._adapter._get_user(), "hpcuser")

    def test_get_file_transfer(self):
        self.assertEqual(
            self.remote._adapter._get_file_transfer(file="abc.txt", local_dir="local", remote_dir="test"),
            os.path.abspath("abc.txt")
        )


import os
from queue import Queue
from concurrent.futures import Future, ThreadPoolExecutor
import unittest

from pysqa.executor.backend import execute_files_from_list
from pysqa.executor.executor import Executor
from pysqa.executor.helper import (
    read_from_file,
    deserialize,
    serialize_funct,
    write_to_file,
    serialize_result,
    reload_previous_futures,
    _get_file_name,
    _set_future,
    _update_task_dict,
)
from pysqa.queueadapter import QueueAdapter


def funct_add(a, b):
    return a+b


@unittest.skipIf(os.name == 'nt', "Runs forever on Windows")
class TestExecutorHelper(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "cache")
        os.makedirs(self.test_dir, exist_ok=True)

    def tearDown(self):
        for f in os.listdir(self.test_dir):
            os.remove(os.path.join(self.test_dir, f))
        os.removedirs(self.test_dir)

    def test_cache(self):
        funct_dict = serialize_funct(fn=funct_add, a=1, b=2)
        file_name_in = write_to_file(
            funct_dict=funct_dict,
            state="in",
            cache_directory=self.test_dir
        )[0]
        self.assertEqual(len(os.listdir(self.test_dir)), 1)
        funct_dict = read_from_file(
            file_name=os.path.join(self.test_dir, file_name_in)
        )
        apply_dict = deserialize(funct_dict=funct_dict)
        key = list(apply_dict.keys())[0]
        v = apply_dict[key]
        result_dict = {key: v["fn"].__call__(*v["args"], **v["kwargs"])}
        file_name_out = write_to_file(
            funct_dict=serialize_result(result_dict=result_dict),
            state="out",
            cache_directory=self.test_dir,
        )[0]
        self.assertEqual(len(os.listdir(self.test_dir)), 2)
        f = Future()
        _set_future(
            file_name=os.path.join(self.test_dir, file_name_out), 
            future=f
        )
        self.assertEqual(f.result(), 3)
        task_dict = {key: Future()}
        _update_task_dict(
            task_dict=task_dict,
            task_memory_dict={},
            cache_directory=self.test_dir
        )
        self.assertEqual(task_dict[key].result(), 3)

    def test_reload_previous_future(self):
        funct_dict = serialize_funct(fn=funct_add, a=1, b=2)
        file_name_in = write_to_file(
            funct_dict=funct_dict,
            state="in",
            cache_directory=self.test_dir
        )[0]
        queue = Queue()
        future_dict_one = {}
        reload_previous_futures(
            future_queue=queue,
            future_dict=future_dict_one,
            cache_directory=self.test_dir
        )
        self.assertEqual(len(future_dict_one), 1)
        self.assertEqual(list(future_dict_one.keys())[0], file_name_in.split(".in.pl")[0])
        self.assertEqual(len(os.listdir(self.test_dir)), 1)
        with ThreadPoolExecutor() as exe:
            execute_files_from_list(
                tasks_in_progress_dict={},
                cache_directory=self.test_dir,
                executor=exe
            )
        self.assertEqual(len(os.listdir(self.test_dir)), 2)
        future_dict_two = {}
        reload_previous_futures(
            future_queue=queue,
            future_dict=future_dict_two,
            cache_directory=self.test_dir
        )
        key = list(future_dict_two.keys())[0]
        self.assertEqual(len(future_dict_two), 1)
        self.assertEqual(key, file_name_in.split(".in.pl")[0])
        self.assertEqual(future_dict_two[key].result(), 3)


class TestExecutor(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "executor_cache")
        os.makedirs(self.test_dir, exist_ok=True)

    def tearDown(self):
        for f in os.listdir(self.test_dir):
            os.remove(os.path.join(self.test_dir, f))
        os.removedirs(self.test_dir)

    def test_executor(self):
        def execute_command(
                commands,
                working_directory=None,
                split_output=True,
                shell=False,
                error_filename="pysqa.err",
        ):
            return str(1)

        queue_adapter = QueueAdapter(
            directory=os.path.join(self.test_dir, "../config/slurm"),
            execute_command=execute_command
        )
        with Executor(
            cwd=self.test_dir,
            queue_adapter=queue_adapter,
            queue_adapter_kwargs={
                "queue": "slurm",
                "job_name": "test",
                "cores": 1
            }
        ) as exe:
            fs = exe.submit(fn=funct_add, a=1, b=2)
            funct_dict = serialize_funct(fn=funct_add, a=1, b=2)
            file_name_in = _get_file_name(name=list(funct_dict.keys())[0], state="in")
            funct_dict = read_from_file(
                file_name=os.path.join(self.test_dir, file_name_in)
            )
            apply_dict = deserialize(funct_dict=funct_dict)
            result_dict = {
                k: v["fn"].__call__(*v["args"], **v["kwargs"]) for k, v in apply_dict.items()
            }
            _ = write_to_file(
                funct_dict=serialize_result(result_dict=result_dict),
                state="out",
                cache_directory=self.test_dir,
            )[0]
            self.assertEqual(fs.result(), 3)



import os
import re
import queue
from concurrent.futures import Future

import hashlib
import cloudpickle


def deserialize(funct_dict):
    try:
        return {k: cloudpickle.loads(v) for k, v in funct_dict.items()}
    except EOFError:
        return {}


def find_executed_tasks(future_queue, cache_directory):
    task_memory_dict = {}
    while True:
        task_dict = {}
        try:
            task_dict = future_queue.get_nowait()
        except queue.Empty:
            pass
        if "shutdown" in task_dict.keys() and task_dict["shutdown"]:
            break
        else:
            _update_task_dict(
                task_dict=task_dict,
                task_memory_dict=task_memory_dict,
                cache_directory=cache_directory,
            )


def read_from_file(file_name):
    name = file_name.split("/")[-1].split(".")[0]
    with open(file_name, "rb") as f:
        return {name: f.read()}


def reload_previous_futures(future_queue, future_dict, cache_directory):
    file_lst = os.listdir(cache_directory)
    for f in file_lst:
        if f.endswith(".in.pl"):
            key = f.split(".in.pl")[0]
            future_dict[key] = Future()
            file_name_out = key + ".out.pl"
            if file_name_out in file_lst:
                _set_future(
                    file_name=os.path.join(cache_directory, file_name_out),
                    future=future_dict[key],
                )
            else:
                future_queue.put({key: future_dict[key]})


def serialize_result(result_dict):
    return {k: cloudpickle.dumps(v) for k, v in result_dict.items()}


def serialize_funct(fn, *args, **kwargs):
    binary = cloudpickle.dumps({"fn": fn, "args": args, "kwargs": kwargs})
    return {fn.__name__ + _get_hash(binary=binary): binary}


def write_to_file(funct_dict, state, cache_directory):
    file_name_lst = []
    for k, v in funct_dict.items():
        file_name = _get_file_name(name=k, state=state)
        file_name_lst.append(file_name)
        with open(os.path.join(cache_directory, file_name), "wb") as f:
            f.write(v)
    return file_name_lst


def _get_file_name(name, state):
    return name + "." + state + ".pl"


def _get_hash(binary):
    # Remove specification of jupyter kernel from hash to be deterministic
    binary_no_ipykernel = re.sub(b"(?<=/ipykernel_)(.*)(?=/)", b"", binary)
    return str(hashlib.md5(binary_no_ipykernel).hexdigest())


def _set_future(file_name, future):
    values = deserialize(funct_dict=read_from_file(file_name=file_name)).values()
    if len(values) == 1:
        future.set_result(list(values)[0])


def _update_task_dict(task_dict, task_memory_dict, cache_directory):
    file_lst = os.listdir(cache_directory)
    for key, future in task_dict.items():
        task_memory_dict[key] = future
    for key, future in task_memory_dict.items():
        file_name_out = _get_file_name(name=key, state="out")
        if not future.done() and file_name_out in file_lst:
            _set_future(
                file_name=os.path.join(cache_directory, file_name_out),
                future=future,
            )



import os
import queue
from concurrent.futures import Future, Executor as FutureExecutor

from pympipool.shared import cancel_items_in_queue, RaisingThread
from pysqa.executor.helper import (
    reload_previous_futures,
    find_executed_tasks,
    serialize_funct,
    write_to_file,
)


class Executor(FutureExecutor):
    def __init__(self, cwd=None, queue_adapter=None, queue_adapter_kwargs=None):
        self._task_queue = queue.Queue()
        self._memory_dict = {}
        self._cache_directory = os.path.abspath(os.path.expanduser(cwd))
        self._queue_adapter = queue_adapter
        reload_previous_futures(
            future_queue=self._task_queue,
            future_dict=self._memory_dict,
            cache_directory=self._cache_directory,
        )
        command = (
            "python -m pysqa.executor --cores "
            + str(queue_adapter_kwargs["cores"])
            + " --path "
            + str(self._cache_directory)
        )
        self._queue_id = self._queue_adapter.submit_job(
            working_directory=self._cache_directory,
            command=command,
            **queue_adapter_kwargs,
        )
        self._process = RaisingThread(
            target=find_executed_tasks,
            kwargs={
                "future_queue": self._task_queue,
                "cache_directory": self._cache_directory,
            },
        )
        self._process.start()

    def submit(self, fn, *args, **kwargs):
        funct_dict = serialize_funct(fn, *args, **kwargs)
        key = list(funct_dict.keys())[0]
        if key not in self._memory_dict.keys():
            self._memory_dict[key] = Future()
            _ = write_to_file(
                funct_dict=funct_dict, state="in", cache_directory=self._cache_directory
            )[0]
            self._task_queue.put({key: self._memory_dict[key]})
        return self._memory_dict[key]

    def shutdown(self, wait=True, *, cancel_futures=False):
        if cancel_futures:
            cancel_items_in_queue(que=self._task_queue)
        self._task_queue.put({"shutdown": True, "wait": wait})
        self._queue_adapter.delete_job(process_id=self._queue_id)
        self._process.join()
